//  ComputationalAssemblerBasic.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;

import java.io.*;
import java.util.*;

import java.lang.reflect.Method;

import swarm.collections.ArrayImpl;

/**
 * The ComputationalAssemblerBasic class instances make 
 * computational processes; we have a computationa assembler instance for
 * each unit
 *
 * THIS CLASS IS NOT USED DIRECTLY, BUT VIA THE INHERITING CLASS
 * ComputationalAssembler.java
 *
 * @author Pietro Terna; Carlo Desotgiu made the steps 1001 and 1002; Alessandro
 * Balla and Roberto Crosetto discovered a severe bug in 1002
 */
public class ComputationalAssemblerBasic extends SwarmObjectImpl{

    /** the memory matrixes used in computations; we can add a lot of these */
    MemoryMatrix mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7, mm8, mm9;
    /** the current layer */
    public int layer;
    /** the optional addendum fpr the production phases copying units */
    public int addendum;

    /** the unit for which we are assembling for */
    public Unit myUnit;
    /** whichUnit is used to set the unit the order is in from outside
     *  (look to 1110 as an example) */
    public Unit whichUnit;
    /** the list of the order requiring computations */
    public ListImpl waitingList;
    /** a computational specification set */
    ComputationalSpecificationSet pendingComputationalSpecificationSet;
    /** a flag to avoid the control of the number of matrixes
     *  this flag has to be used from a computational step using an existing step
     *  and modifying on the fly the memory matrix number of the used computational
     *  step  */
    public boolean checkMatrixNumber;
    /** the status of a specific computation */
    public boolean done;

    /** internal displacements of rows and cols to be used when invoking the methods
     *  of the BASIC computational steps via other computational steps <br><br>
     *
     *  REMEMBER TO RESET ALWAYS TO ZERO THE DISPLACEMENTS AFTER THE USE<br><br>
     *
     * */
    public int rd, cd,   // memory matrixes
	       urd, ucd, urd0, ucd0, urd1, ucd1; // unit memory matrixes

    // object to be addressed
    OrderDistiller myOrderDistiller;
    ESFrameModelSwarm eSFrameModelSwarm;

    /**
     * the constructor for ComputationalAssembler
     */
    public ComputationalAssemblerBasic (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone);

	init(aZone, od, mo);
    }

    /** managing the constructor content */
    public void init (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {

	myOrderDistiller = od;
	eSFrameModelSwarm=mo;

	// the list of orders with computational processes to be assembled
	waitingList = new ListImpl (getZone());

	// the diplacements
	rd=0; cd=0;
	urd=0; ucd=0; urd0=0; ucd0=0; urd1=0; ucd1=0;
	// the check of the matrix number
	checkMatrixNumber=true;
	// no outside setting for the unit the order is in
	whichUnit=null;
	// the optional addendum fpr the production phases copying units
	addendum=0;

    }


    /** setting the unit we are assembling for */
    public void setUnit(Unit u){
	myUnit = u;
    }

    /** adding an order to the waitingList */
    public boolean setComputationalWaitingList(Order anOrder){
	waitingList.addLast(anOrder);
	return true;
    }

    /** making computations and eliminating the orders from our
     *  waitingList */
    public void checkingComputationsAndFreeingOrders(){
	int n, nmax, t;
	Order anOrder;

	nmax = waitingList.getCount();
	if(nmax>0)
	    for (n=0;n<nmax;n++)
		{
		    anOrder = (Order) waitingList.removeFirst();

		    pendingComputationalSpecificationSet = 
			(ComputationalSpecificationSet)
			 anOrder.getPendingComputationalSpecificationSet();

		    if(pendingComputationalSpecificationSet == null){
			System.out.println("Internal error in" +
				       " computationalAssembler of unit " +
				       myUnit.getUnitNumber() +
				       " - order # " + 
				       anOrder.getOrderNumber() +  
				       "waiting for computations "+
				       "has no computational specifications.");
			MyExit.exit(1);
		    }

		    // computations
		    done=false;
		    if(StartESFrame.verbose)
				System.out.println
				    ("Making computations for order # " +
				     pendingComputationalSpecificationSet.
				     getOrderNumber() + " of type " +
				     pendingComputationalSpecificationSet.
				     getComputationType() + " with " + 
				     pendingComputationalSpecificationSet.
				     getNumberOfMemoryMatrixesToBeUsed() +
				     " matrixes");

				t=pendingComputationalSpecificationSet.
				    getComputationType();
				// transforming the code contained in t (a
				// number in the range -1000 -1999) in a 
				// method label (c1000 c1999) via reflection
				// methods
				try{
				    Class c = this.getClass();
				    Method m = c.getMethod("c"+(-1*t),null);
				    m.invoke(this, null);
				}
				catch (Exception e) {
				    System.out.println(e);
				    System.out.println("Invalid computational"+
						       " code in order: " +
						       anOrder.
						       getOrderNumber());
				    MyExit.exit(2);
				}


		    // exiting
		    if(done)
			{
			    if(StartESFrame.verbose)
				System.out.println
				    ("Made computations for order # " +
				     pendingComputationalSpecificationSet.
				     getOrderNumber());
			    pendingComputationalSpecificationSet.setDone();
			    // the last set at present has not effect
			    // it is introduced for possible future uses
			}

		    else waitingList.addLast(anOrder);

		}
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }

    /**
     * checking if an order is in the waitingList
     */
    public boolean thisOrderIsInTheWaitingList(Order o){

	// the result is true if o is a member of waitingList; false in
	// the opposite case
	return waitingList.contains(o);
    }

    /** removing an order form the waintingList */
    public void removeThisOrderFromTheWaitingList(Order o){
	waitingList.remove(o);
    }

     /** computational operations with code -1995 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a 1 in position 1,1 of the
     *  unique received matrix and set the status to done
     */
    public void c1995(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1995 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,1+rd,1+cd,1.0);
	mm0.print();

	done=true;
    } // end c1995

    /** computational operations with code -1996 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a 1 in position 0,0 of the
     *  unique received matrix and set the status to done
     */
    public void c1996(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1996 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,0+rd,0+cd,1.0);
	mm0.print();

	done=true;
    } // end c1996

    /** computational operations with code -1997 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a 1 in position 1,0 of the
     *  unique received matrix and set the status to done
     */
    public void c1997(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1997 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,1+rd,0+cd,1.0);
	mm0.print();

	done=true;
    } // end c1997

    /** computational operations with code -1998 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a number in position 0,0 of the
     *  unique received matrix and set the status to done
     */
    public void c1998(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1998 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,0+rd,0+cd,1.0);
	mm0.print();

	done=true;
    } // end c1998

    /** computational operations with code -1999 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code verifies position 0,0 of the three
     *  received matrixes; only if these positions are all not empty
     *  the code empties them and set the status to done
     */
    public void c1999(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=3 && checkMatrixNumber)
	    {
		System.out.println("Code -1999 requires three matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	mm2=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(2);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	if(! (mm0.getEmpty(layer,0+rd,0+cd) || mm1.getEmpty(layer,0+rd,0+cd) 
	                              || mm2.getEmpty(layer,0+rd,0+cd) ) )
	    {
		mm0.setEmpty(layer,0+rd,0+cd);
		mm1.setEmpty(layer,0+rd,0+cd);
		mm2.setEmpty(layer,0+rd,0+cd);
		done=true;
	    }

    } // end c1999


    // computational steps 1001 and 1002 to launch recipes from recipes

    /**
     * computational operations with code -1001
     *
     * this computational code places the numbers 
     * of the recipes to launch from other recipes, via the c step 1002,
     * in position 0,0 of the received matrixes and set the status to done.
     *
     * the number of the matrixes used here must be less or equal to totalMemoryMatrixNumber
     * (the first m matrixes of these ones); beside this, m has to be less than 10
     */  
    public void c1001()
    {
	MemoryMatrix mm;

	int m, i;
	
	int[] recipesToLaunch;
	recipesToLaunch = new int[10]; 

	// from pending specifications related to the recipe launching this computational step
	m = pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed();
	if(m > eSFrameModelSwarm.getTotalMemoryMatrixNumber()
	   || m > 10)
	    {
		System.out.println("The number of types of the recipes launched by recipes\n"+
				   "cannot exceed 10 or the totalMemoryMatrixNumber.");
		MyExit.exit(1);
	    }

	try
	{    
	    BufferedReader fileRecipesFromRecipes = null;
	    StringTokenizer t;
	    String line = " "; 

	    fileRecipesFromRecipes = new BufferedReader(new 
				      FileReader("recipeData0/recipesFromRecipes.txt"));

	    t = new StringTokenizer(line, " ");
	    for (i = 0; i < m; i++)
	    {	
		t=MyReader.check("recipeData0/recipesFromRecipes.txt",
				 fileRecipesFromRecipes, line, t);
		recipesToLaunch[i] = Integer.parseInt(t.nextToken());
	    }

	    fileRecipesFromRecipes.close();
	}    
	catch (FileNotFoundException fnfe)
	{
	    System.out.println(fnfe);
	    MyExit.exit(1);
	} 
	catch (IOException e)
	{
	    System.out.println("IOException:" + e.toString());
	    MyExit.exit(1);
	}

	for (i = 0; i < m; i++)
	{    
	    mm =(MemoryMatrix) pendingComputationalSpecificationSet.
		getMemoryMatrixAddress(i);
	    layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	    mm.setValue(layer,0+rd,0+cd,recipesToLaunch[i]);
	}
	done=true;
    } // end c1001



    public void c1002()
    {
	MemoryMatrix mm;
	int aRecipe;
	
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1002 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm =(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();
	
	aRecipe =(int) mm.getValue(layer,0+rd,0+cd);
	myOrderDistiller.setComputationalRecipe(aRecipe); // a recipe to be launched
	                                                  // at the end of this cycle
	
	done=true;
    } // end c1002


    /**
     * computational operations with code -1100  (a code for the jES Open Foundation
     *  extension) <br><br>
     *
     * this computational step acts as generalized matrix loader that places the values 
     * found in unitData0/memoryMatrixContents.txt file in the matrixex and rows and cols
     * reported in the same file
     *
     * the number of the matrixes used here must be less or equal to totalMemoryMatrixNumber
     * (the first m matrixes of these ones); beside this, m has to be less than 10
     *
     * it is mandatory to enter all the defined matrixes 
     * in their original in the c 1100 m m1 m2 ... command
     *
     * values in the unitData0/memoryMatrixContents.txt can be in any order
     *
     * finally it changes the status to done;     */  
    public void c1100()
    {
	ArrayImpl mm;
	int m, i, mi, la, r, c;
	float value;
	
	// number of matrixes to deal with
	m = pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed();
	if(m != eSFrameModelSwarm.getTotalMemoryMatrixNumber()
	   || m > 10)
	    {
		System.out.println("The number of matrix in computaional step 1100\n"+
				   "must be equal to totalMemoryMatrixNumber;\n"+
				   "anyway, it cannot exceed 10.");
		MyExit.exit(2);
	    }

	mm = new ArrayImpl(getZone(), m);

	for (i = 0; i < m; i++)
	    {
		if(((MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(i)).
		   getMemoryMatrixNumber()!=i)
		    {
		System.out.println("In computational step 1100 the sequence of the numbers\n"+
				   " of the matrixes must be in the creation order, as in file \n"+
				   "unitData0/memoryMatrixes.txt");
		MyExit.exit(2);
		    }

	    mm.atOffset$put(i, pendingComputationalSpecificationSet.
			    getMemoryMatrixAddress(i));
	    }

	// read the matrix numbers and the layer, row, col and value information
	// from the unitData0/memoryMatrixContents.txt file
	try{
	    // read the file
	    BufferedReader fileMatrixValues = null;
	    StringTokenizer t;
	    fileMatrixValues = new 
		BufferedReader(new FileReader("unitData0/memoryMatrixContents.txt"));

	    // reading a control line (titles)
	    String line = " ", lineTest = 
		"number_layer(always_0_in_jESOF)_row_col_value";

	    line=fileMatrixValues.readLine();
	    if (! line.equals(lineTest))
		{
		    System.out.println(
		    "First line in unitData0/memoryMatrixContents.txt must contain:");
		    System.out.println(lineTest);
		    MyExit.exit(1);
		}

	    line=" ";
	    // starting with a void line
	    t = new StringTokenizer(line, " ");

	    t=MyReader.check_with_eof("unitData0/memoryMatrixContents.txt", 
			     fileMatrixValues, line, t);
	    while(t!=null)
		{
		    mi = Integer.parseInt(t.nextToken());

		    t=MyReader.check_with_eof("unitData0/memoryMatrixContents.txt", 
				     fileMatrixValues, line, t);
		    la = Integer.parseInt(t.nextToken());

		    t=MyReader.check_with_eof("unitData0/memoryMatrixContents.txt", 
				     fileMatrixValues, line, t);
		    r  = Integer.parseInt(t.nextToken());

		    t=MyReader.check_with_eof("unitData0/memoryMatrixContents.txt", 
				     fileMatrixValues, line, t);
		    c  = Integer.parseInt(t.nextToken());

		    t=MyReader.check_with_eof("unitData0/memoryMatrixContents.txt", 
				     fileMatrixValues, line, t);
		    value = Float.parseFloat(t.nextToken());
		    if(mi>m-1)
			{
			    System.out.println("Attempt to use a not existent matrix, i.e. "+mi);
			    MyExit.exit(2);
			}
		    if(la!=0)
			{
			    System.out.println("Layer must be 0 in reference to matrix "+mi);
			    MyExit.exit(2);
			}
		    if(r>=((MemoryMatrix) mm.atOffset(mi)).getNumberOfRows())
			{
			    System.out.println("Attempt to go beyond the max number of rows"+
					       " in reference to matrix "+mi);
			    MyExit.exit(2);
			}
		    if(c>=((MemoryMatrix) mm.atOffset(mi)).getNumberOfCols())
			{
			    System.out.println("Attempt to go beyond the max number of cols"+
					       " in reference to matrix "+mi);
			    MyExit.exit(2);
			}
		   		    
		    ((MemoryMatrix) mm.atOffset(mi)).setValue(la,r,c,value);

		    t=MyReader.check_with_eof("unitData0/memoryMatrixContents.txt", 
				     fileMatrixValues, line, t);
		}
	    fileMatrixValues.close();
	}
	catch (FileNotFoundException fnfe)
	    {
		System.out.println(fnfe);
		MyExit.exit(1);
	    } 
	catch (IOException e)
	    {
		System.out.println("IOException:" + e.toString());
		MyExit.exit(1);
	    }
	
	for(i=0;i<m;i++)((MemoryMatrix) mm.atOffset(i)).print();

	done=true;
    } // end c1100

    /** computational operations with code -1101 (a code for the jES Open Foundation
     *  extension) <br><br>
     *
     *  this computational code creates a new random unit in the model stratum 
     *  reported in position 0,1 of the unique received matrix, with the probability
     *  set in position 0,0 of the same matrix; it chose randomly a unit in the 
     *  proforma unit list, generated from the stratum unitBasicData.txt file, from the
     *  top to the 'potentialUnitTypes-1' row; finally it changes the 
     *  status to done
     *
     *  it returns the new unit address, to be used, if necessary, in 
     *  ComputationalAssembler.java
     *
     */
    public Unit c1101(){
	int stratum; 
	double prob;
	Unit aUnit;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1101 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	prob    = mm0.getValue(layer,0+rd,0+cd);
	stratum = (int) mm0.getValue(layer,0+rd,1+cd);

	if(stratum<0 || stratum>=InformationRepository.getStratumNumber())
	    {
		System.out.println("Error in computational step 1101;"+
				   " the stratum number cannot be accepted");
		MyExit.exit(2);
	    }

	aUnit=null;

	if(prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    aUnit=((ESFrameModelSwarm) InformationRepository.getESFrameModelSwarm(stratum)).
		generateANewUnitFromAComputationalStep();

	done=true;
	return aUnit;
    } // end c1101


    /** computational operations with code -1102 (a code for the jES Open Foundation
     *  extension) <br><br>
     *
     *  this computational code drops the unit the order is in, with the probability
     *  set in position 0,0 of the first received matrix and increases the 
     *  position 0,1 of the second received matrix by 1, to count the dropped
     *  units; this computational step must be in the last step of the order, because
     *  the dropped unit cannot send it to the successive one; the computational step,
     *  before setting the unit to be dropped, applies the orderDoneStep() method to 
     *  the order, for accounting purposes; finally it changes the status to done
     */
    public void c1102(){
	double prob;
	float count;
	Unit theUnitToBeDropped;
	Order theCurrentOrder;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2 && checkMatrixNumber)
	    {
		System.out.println("Code -1102 requires two matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	prob    = mm0.getValue(layer,0+rd,0+cd);
	if(prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
		theUnitToBeDropped=theCurrentOrder.getTheUnitTheOrderIsIn();
		if(! theUnitToBeDropped.getDroppable()) // if it is droppable
		                                        // the same kind of order
                                                        // acted in this unit in the
                                                        // the same cycle
		    {
			theCurrentOrder.setDoneStep(theUnitToBeDropped.getUnitNumber());
			if(theCurrentOrder.getNextStep()!=0) // the concluded step, 
			                                     // containg the computation,
			                                     // has to be the last one
			    {
				System.out.println("Error in computational step 1102,"+
						   " which has to be the last one.");
				MyExit.exit(1);
			    }
			
			// add the unit to the dropped count, unless it has just been dropped
			// in the same cycle
			if(! theUnitToBeDropped.getDroppable())
			    {
				count=mm1.getValue(layer,0+rd,1+cd);
				mm1.setValue(layer,0+rd,1+cd,++count);
			    }
			// drop the unit (the currentOrder is dropped dropping the unit)
			theUnitToBeDropped.setDroppable();
		    }
	    }

	mm1.print();
	done=true;
    } // end c1102


    /** computational operations with code -1103 (a code for the jES Open Foundation
     *  extension) <br><br>
     *
     *  this computational code creates a copy of the unit the order is in the model 
     *  stratum of the original unit, near to it with the standard placement rule, with 
     *  the probability set in position 0,0 of the first received matrix; 
     *  the count of the created units is reported increasing by one the position 
     *  0+rd,1+cd of the second received matrix
     *  finally it changes the status to done
     */
    public void c1103(){
	double prob, count;
	Unit theUnitTheOrderIsIn, aUnit;
	Order theCurrentOrder;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2 && checkMatrixNumber)
	    {
		System.out.println("Code -1103 requires two matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	prob    = mm0.getValue(layer,0+rd,0+cd);

	if(prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
		theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
		aUnit=((ESFrameModelSwarm) theUnitTheOrderIsIn.getUnitModel()).
		    copyUnitNearTheOriginalPosition(theUnitTheOrderIsIn,(int) 0); 
		                                           // aUnit is used below
                                                           // if aUnit==null no space
                                                           // has been found and the unit
                                                           // has not been created

		if(aUnit != null)
		    {
			count=mm1.getValue(layer,0+rd,1+cd);
			mm1.setValue(layer,0+rd,1+cd,++count);
		    }

		if(StartESFrame.verbose)
		    if(aUnit==null)System.out.println("Order n. " +theCurrentOrder.getOrderNumber()+
					    " failed creating a new unit" +
					    " at X="+aUnit.getXPos()+" and Y="+aUnit.getXPos()+
					    " in stratum "+((ESFrameModelSwarm) aUnit.getUnitModel()).
					    getStratum());
		    else System.out.println("Order n. " +theCurrentOrder.getOrderNumber()+
					    " has created the new unit n." + aUnit.getUnitNumber()+
					    " at X="+aUnit.getXPos()+" and Y="+aUnit.getXPos()+
					    " in stratum "+((ESFrameModelSwarm) aUnit.getUnitModel()).
					    getStratum());
		
	    }

	done=true;
    } // end c1103

    /** computational operations with code -1104 (a code for the jES Open Foundation
     *  extension) <br><br>
     *
     *  this computational code creates a copy of the unit the order is in, in the model 
     *  stratum of the original unit, in a random position with the standard placement rule, with 
     *  the probability set in position 0,0 of the first received matrix; 
     *  the count of the created units is reported increasing by one the position 
     *  0+rd,1+cd of the second received matrix
     *  finally it changes the status to done
     */
    public void c1104(){
	double prob, count;
	Unit theUnitTheOrderIsIn, aUnit;
	Order theCurrentOrder;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2 && checkMatrixNumber)
	    {
		System.out.println("Code -1104 requires two matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	prob    = mm0.getValue(layer,0+rd,0+cd);

	if(prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
		theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
		aUnit=((ESFrameModelSwarm) theUnitTheOrderIsIn.getUnitModel()).
		    copyUnitInARandomPosition(theUnitTheOrderIsIn,(int) 0); 
		                                           // aUnit is used below
                                                           // if aUnit==null no space
                                                           // has been found and the unit
                                                           // has not been created

		if(aUnit != null)
		    {
			count=mm1.getValue(layer,0+rd,1+cd);
			mm1.setValue(layer,0+rd,1+cd,++count);
		    }

		if(StartESFrame.verbose)
		    if(aUnit==null)System.out.println("Order n. " +theCurrentOrder.getOrderNumber()+
					    " failed creating a new unit" +
					    " at X="+aUnit.getXPos()+" and Y="+aUnit.getXPos()+
					    " in stratum "+((ESFrameModelSwarm) aUnit.getUnitModel()).
					    getStratum());
		    else System.out.println("Order n. " +theCurrentOrder.getOrderNumber()+
					    " has created the new unit n." + aUnit.getUnitNumber()+
					    " at X="+aUnit.getXPos()+" and Y="+aUnit.getXPos()+
					    " in stratum "+((ESFrameModelSwarm) aUnit.getUnitModel()).
					    getStratum());
		
	    }

	done=true;
    } // end c1104

    /** computational operations with code -1110 (a code for the jES Open Foundation
     *  extension) <br><br>
     *
     *  this computational code creates a copy of the unit the order is in, in the same x,y place 
     *  of the copied unit, but in another model stratum, with the probability set in position 0,0 
     *  of the unique received matrix; the created unit is set in the stratum indicated in position 
     *  0,1 of the unique received matrix; the unit is created, with the above probability, if the 
     *  content of the unit memory matrix in position 0,0 is positive (the actual position can be 
     *  modified via the calling step, i.e. 1210 in tutorial step3c); if the destination position 
     *  in the new stratum is occupied the new unit is not created, without any advice; the 
     *  new unit is not created if the original one is in the destination stratum; the new unit 
     *  has the same visibility of the old one; 
     *  finally, the computational step changes the status to done; 
     *
     *  c1110 is not normally utilized directly but via a special computational step (i.e. 1210 
     *  in tutorial step3c) setting the position and content of the cell (<=0 or >0) indicated above; 
     *
     *  the memory matrix of the copied unit is shared among the original unit itself and its copies,
     *  avoiding the creation of new matrixes for the copies    
     *
     */
    public Unit c1110(){
	double prob, count;
	Unit theUnitTheOrderIsIn, aUnit;
	Order theCurrentOrder;
	MemoryMatrix mU;
	int toStratum;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1110 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	prob    = mm0.getValue(layer,0+rd,0+cd);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();

	// if whichUnit is not null, the unit we are in is set from outside
	if(whichUnit==null)theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	else theUnitTheOrderIsIn=whichUnit;

	mU=theUnitTheOrderIsIn.getMemoryMatrix();

	aUnit=null;
	if(mU.getValue(layer,0+urd,0+ucd)>0 
	   && prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		toStratum = (int) mm0.getValue(layer,0+rd,1+cd);
		// unit not in the destination stratum
		if(toStratum != ((ESFrameModelSwarm) theUnitTheOrderIsIn.getUnitModel()).
		   getStratum())
		    {

			aUnit=((ESFrameModelSwarm) theUnitTheOrderIsIn.getUnitModel()).
			    copyUnitToAnotherStratum(theUnitTheOrderIsIn,(int) 0, addendum, 
						     toStratum); // aUnit is used below
			                                         // if aUnit==null no space
                                                                 // has been found and the unit
                                                                 // has not been created

		    }
	    }

	done=true;
	return aUnit;
    } // end c1110

    /** computational operations with code -1111 (a code for the jES Open Foundation
     *  extension) <br><br> 
     *
     *  this step acts exactly as c1110, but
     */
    public Unit c1111(){
	Unit theUnitTheOrderIsIn, aUnit;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1111 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	addendum = (int) mm0.getValue(layer,0+rd,2+cd);
	aUnit=c1110();

	addendum=0;

	// done = true in c1110
	return aUnit;
    } // end c1111

    /** computational operations with code -1120 (a code for the jES Open Foundation
     *  extension) <br><br> 
     *
     *  this computational code modifies the position 0,0 both of the memory matrix 
     *  of the unit the order is in and of the memory matrix of the unit the order 
     *  come from; the amount of the two modifications is defined by two addenda stored 
     *  in position 0,0 and 0,1 of the unique received matrix (the actual position can 
     *  be modified via the calling computational step, of the class 12xx); also the 
     *  position of the two memory matrixes can be modified via the via the 12xx calling 
     *  computational step; in 0,0 the addendum 1 for the unit the order is in and 
     *  addendum 2 for the unit the order come from; finally it changes the  status 
     *  to done;
     */
    public void c1120(){
	float addendum1, addendum2;
	Order theCurrentOrder;
	Unit theUnitTheOrderIsIn, theUnitTheOrderComesFrom;
	MemoryMatrix mU1, mU2;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1120 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	addendum1 = mm0.getValue(layer,0+rd,0+cd);
	addendum2 = mm0.getValue(layer,0+rd,1+cd);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	theUnitTheOrderComesFrom=theCurrentOrder.getTheUnitTheOrderComesFrom();

	mU1=theUnitTheOrderIsIn.getMemoryMatrix();
	mU2=theUnitTheOrderComesFrom.getMemoryMatrix();

	if(! theUnitTheOrderIsIn.getDroppable())
	    {
		mU1.setValue(layer, 0+urd0, 0+ucd0, mU1.
			     getValue(layer,0+urd0, 0+ucd0)+addendum1);
		mU2.setValue(layer, 0+urd1, 0+ucd1, mU2.
			     getValue(layer,0+urd1, 0+ucd1)+addendum2);
	    }

	done = true;
    } // end c1120

    /** computational operations with code -1199 (a code for the jES Open Foundation
     *  extension) <br><br> 
     *
     *  this highly usable computational code, when received by one unit in a stratum, 
     *  acts vs. all the units of the same type (i.e., with the same production 
     *  phase) in all strata; actions: with the probability 
     *  set in position 0,0 of the first received matrix, c1999 adds the value contained 
     *  in 0,5 of the second received matrix to each unit memory matrix a row and column
     *  reported in pos 0,3 and 0,4 of the second matrix; if the resulting content is < the 
     *  level contained in position 0,2 of the second matrix, the considered unit is 
     *  dropped (if it is the unit the order is in, also the order is dropped), counting 
     *  the dropped units in position 0,1 of the second matrix (obviously, first and second 
     *  matrixes can be coincident)
     *  finally it changes the status to done
     */
    public void c1199(){
	double prob;
	float delta, level, energy, count;
	int unitCount, layer=0, row, col, i, im, thePhase;
	Unit theUnitTheOrderIsIn, aUnit;
	Order theCurrentOrder;
	ListImpl aList;
	MemoryMatrix mU;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2 && checkMatrixNumber)
	    {
		System.out.println("Code -1199 requires two matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	prob    = mm0.getValue(layer,0+rd,0+cd);

	if(prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
		    getMemoryMatrixAddress(1);

		theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
		theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
		thePhase=theUnitTheOrderIsIn.getProductionPhase();

		row=(int) mm1.getValue(layer,0+rd,3+cd);
		col=(int) mm1.getValue(layer,0+rd,4+cd);
		level=mm1.getValue(layer,0+rd,2+cd);
		delta=mm1.getValue(layer,0+rd,5+cd);

		for (im=0;im<InformationRepository.getStratumNumber();im++)
		    {
			aList=(ListImpl) InformationRepository.getUnitListFromModelSequence(im);
			unitCount=aList.getCount();
			for (i=0; i<unitCount;i++)
			    {
				aUnit=(Unit) aList.atOffset(i);
				// a unit showing the same production phase of the
				// receiving unit
				if(thePhase==aUnit.getProductionPhase())
				    {
					mU=aUnit.getMemoryMatrix();
					energy=mU.getValue(layer,row,col);
					energy+=delta;
					if(! aUnit.getDroppable())
					    mU.setValue(layer,row,col,energy); // new energy value

					
					if(energy<level)
					    {
						if(aUnit==theUnitTheOrderIsIn)
						    theCurrentOrder.
							setDoneStep(theUnitTheOrderIsIn.getUnitNumber());
						
						// add the unit to the dropped count, 
						// unless it has just been dropped
						// in the same cycle
						if(! aUnit.getDroppable())
						    {
							count=mm1.getValue(layer,0+rd,1+cd);
							mm1.setValue(layer,0+rd,1+cd,++count);
						    }
						aUnit.setDroppable();
					    }
				    }
			    }
		    }

	    }

	done=true;

    } // end c1199

}



